/* 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <string>
#include <fstream>
#include <sstream>
#include <assert.h>
#include <algorithm>
#include "math.h"
#include "h/mpg.h"
#include "h/conf.h"

using namespace std;

/*Constructor*/
MeanPayoffGame::MeanPayoffGame(const char* filename){
	load(filename);
	assert(is_well_defined());
}

/*Another Constructor*/
MeanPayoffGame::MeanPayoffGame(unsigned int n_0, unsigned int n_1){
	assert(n_0+n_1>0);
	this->n_0 = n_0; // black nodes
	this->n_1 = n_1; // black nodes
	init_arcs();
}

/*Destructor*/
MeanPayoffGame::~MeanPayoffGame(){
	if(VERBOSE_MODE) o_stream << "Destroying MPG... ";
	delete_arcs();
	if(VERBOSE_MODE) o_stream << "done!"<< endl;
}

/* Solves Decision-MPG problem executing the procedure of Zwick and Paterson */
bool* MeanPayoffGame::ZP_solve_decision(){
	if(VERBOSE_MODE) o_stream << "Solving Decision-MPG by Zwick & Paterson procedure ..." << endl;
	unsigned int size = this->n_0 + this->n_1; // wlog > 0
	unsigned long W = get_W();
	unsigned long long K = W > 0 ? 4*size*size*W : 1;
	long long* nu_k = compute_nu_k(size, K);
	bool* decision = new bool[size];
	for(unsigned int u=0; u < size; u++){ // for each vertex u, test threshold
		o_stream << nu_k[u] << " ";
		long long threshold = 2*size*nu_k[u] + K; 
		o_stream << threshold << endl;
		if(threshold > 0) 		// player Max wins
			decision[u] = true;
		else 				// player Min wins
			decision[u] = false;
		const char* winner = decision[u] ? "Max" : "Min";
		if(VERBOSE_MODE) o_stream << "\t node #" 
			<< u << " is winning for player " << winner << endl; 
	}
	delete [] nu_k;
	if(VERBOSE_MODE) o_stream << "... done!" << endl << endl;
	return decision;
}

/* Solves Decision-MPG problem executing the classical procedure of Z.P.*/
t_rational* MeanPayoffGame::ZP_solve_opt_values(){
	if(VERBOSE_MODE) o_stream 
		<< "Solving Values-MPG by classic Zwick & Paterson optimal-values procedure ... " << endl;
	unsigned int size = this->n_0 + this->n_1; // wlog > 0
	unsigned long W = get_W();
	unsigned long K = W > 0 ? 4*size*size*size*W : 1; // one "size" factor more
	long long* nu_k = compute_nu_k(size, K); // as in ZP_solve_decision
	t_rational* values = new t_rational[size];
	for(unsigned int u=0; u < size; u++){ // for each vertex u, compute optimal value
		o_stream << nu_k[u] << " ";
		t_rational opt_value = search_for_rational(nu_k[u], K);
		values[u] = opt_value;
	}
	delete [] nu_k;
	if(VERBOSE_MODE) o_stream << "... done!" << endl << endl;
	return values;	
}

long long floor(t_rational q){
	long long fl = q.num/q.denom;
	if((q.num % q.denom > 0) && ((q.num > 0 && q.denom < 0) || (q.num < 0 && q.denom > 0)))
		fl--;		
	return fl;
}

long long ceil(t_rational q){
	long long ce = q.num/q.denom;
	if((q.num % q.denom > 0) && ((q.num > 0 && q.denom > 0) || (q.num < 0 && q.denom < 0)))
		ce++;
	return ce;
}

/* Searches in time O(size)=O(n) for a rational number inside (L,R) interval given by @nu_k */
t_rational MeanPayoffGame::search_for_rational(long long nu_k, unsigned long K){
	unsigned int size = this->n_0 + this->n_1; // wlog > 0
	t_rational ret_q;
	for(unsigned long denom=1; denom <= size; denom++){
		t_rational R; R.num = denom*(2*nu_k*size*(size-1)+K); R.denom = 2*size*(size-1)*K;
		t_rational L; L.num = denom*(2*nu_k*size*(size-1)-K); L.denom = 2*size*(size-1)*K;
		long long floor_R = floor(R), ceil_L = ceil(L);
		if(floor_R == ceil_L){
			ret_q.num = floor_R;
			ret_q.denom = denom;
			return ret_q;
		}
	}	
	throw "rational value not found, something goes wrong";
}

long long* MeanPayoffGame::compute_nu_k(unsigned int size, unsigned long long K){
	unsigned long W = get_W();
	if(VERBOSE_MODE) o_stream << "W = " << W << endl;
	if(VERBOSE_MODE) o_stream << "K = " << K << endl;
	long long* nu_k = new long long[size]; 
	for(unsigned int u=0; u < size; u++)
		nu_k[u] = 0;
	for(unsigned long long k=0; k < K; k++){
		long long *next_nu_k = new long long[size];
		for(unsigned int u=0; u < size; u++){
			list<t_w_arc*>* arc_list = get_arcs(u);
			//init value
			t_w_arc* front_w_arc = arc_list->front();
			long long value = nu_k[front_w_arc->head_idx] + front_w_arc->weight;
			//compute value
			list<t_w_arc*>::iterator it = arc_list->begin();
			for(it; it != arc_list->end(); it++){
				t_w_arc* arc = *it;
				unsigned int v = arc->head_idx;
				long w = arc->weight;
				long long candidate_value = nu_k[v] + w; 
				if((u < this->n_0 && candidate_value < value) || // for Player Min 
				   (u >= this->n_0 && candidate_value > value))  // for Player Max
					value = candidate_value;		
			}
			next_nu_k[u] = value;
		}
		long long *to_del = nu_k;
		nu_k = next_nu_k;
		delete [] to_del;
	}
	return nu_k;
}

/* returns n_0 */
unsigned int MeanPayoffGame::get_n_0(){
	return this->n_0;
}

/* returns n_1 */
unsigned int MeanPayoffGame::get_n_1(){
	return this->n_1;
}

/* returns e */
unsigned int MeanPayoffGame::get_e(){
	return this->e;
}

/* computes and returns the max abs value of arc weights */
unsigned long MeanPayoffGame::get_W(){
	unsigned long W=0;
	for(unsigned int i=0; i < this->n_0 + this->n_1; i++){
		list<t_w_arc*>* arc_list = get_arcs(i);
		list<t_w_arc*>::iterator it = arc_list->begin();
		for(it; it != arc_list->end(); it++){
			t_w_arc* arc = *it;
			long weight = arc->weight;
			if((weight > 0 && weight > W) || (weight < 0 && -weight > W))
				W = weight > 0 ? weight : -weight;
		}
	}
	return W;
}

/* returns arc list for tail index @tail_idx */
list<t_w_arc*>* MeanPayoffGame::get_arcs(unsigned int tail_idx){
	return &this->arcs[tail_idx];
}

/* sets arcs @arc_list for tail @u */
void MeanPayoffGame::set_arcs(unsigned int u, list<t_w_arc*>* arc_list){
	assert(arc_list->size()>0);
	this->e -= this->arcs[u].size();
	this->arcs[u] = *arc_list;
	this->e += this->arcs[u].size();
}

/**
* Initializes @this MPG by loading the input file @filename
* see "dat/mpg.dat" for an MPG file format example.
**/
void MeanPayoffGame::load(const char* filename){
	/*
	* init_steps:
	*	0: read number of vertices
	*	1: read number of arcs 
	* 	2: read arcs  
	*/
	int init_step = 0;
	/**/
	if(VERBOSE_MODE) o_stream << endl << "Verbose mode on"<<endl; else o_stream << endl <<"Verbose mode off"<<endl;
	if(VERBOSE_MODE) o_stream << "Loading Mean Payoff Game from input file: " << filename << "..." << endl;
	ifstream input(filename);
	string line;
	unsigned int arc_counter = 0;
	if(input.is_open()){
		unsigned long long num_line = 0;
		int num_white, num_black, num_arcs;
		while(getline(input, line)){
			num_line++;
			if(line.compare(0,1,"#")==0){ // skip comments
				if(VERBOSE_MODE) o_stream << "   ... comment found at line #" << num_line << endl;
				continue;
			}
			istringstream iss(line);
			switch(init_step){
				case 0:
					if(VERBOSE_MODE) o_stream 
						<< "   ... reading vertex number at line #" << num_line;
					iss >> num_black >> num_white;
					assert(num_black >= 0);
					assert(num_white >= 0);
					this->n_0 = num_black;
					this->n_1 = num_white;
					init_arcs();
					init_step++;
					if(VERBOSE_MODE) o_stream << ": black vertices " << num_black 
							<< "; white vertices " << num_white << endl;
				break;
				case 1:
					if(VERBOSE_MODE) o_stream 
						<< "   ... reading arcs number at line #" << num_line;
					iss >> num_arcs;
					assert(num_arcs >= 0);
					init_step++;
					if(VERBOSE_MODE) o_stream << ": number of arcs " << num_arcs << endl;
				break;
				case 2:
					if(VERBOSE_MODE) o_stream << "   ... reading arc data at line #" << num_line;
					assert(arc_counter < num_arcs);
					int tail, head;
					long weight;
					iss >> tail >> head >> weight; // assume indexes are absolute 
					assert(tail >= 0 && tail < this->n_0 + this->n_1);
					assert(head >= 0 && head < this->n_0 + this->n_1);
					assert(head != tail); // we assume no self loop  
					t_w_arc* new_arc = new t_w_arc;
					new_arc->head_idx = head;
					new_arc->weight = weight;
					push_arc(tail, new_arc);
					arc_counter++;
					if(VERBOSE_MODE) o_stream << ": tail " << tail 
					     << " head " << head 
					     <<" weight " << weight << endl;
				break;
			}
		}
		input.close();
	}else throw "cannot open input file";
	if(VERBOSE_MODE) o_stream << "... Mean Payoff Game load completed!" << endl<<endl;
}

/* initializes arcs array list */
void MeanPayoffGame::init_arcs(){
	unsigned int size = this->n_0 + this->n_1;
	this->arcs = new list<t_w_arc*>[size];
	this->e=0;
}

/* pushes #arc into tail with index tail_idx*/
void MeanPayoffGame::push_arc(unsigned int tail_idx, t_w_arc* arc){
	assert(tail_idx < this->n_0 + this->n_1);
	get_arcs(tail_idx)->push_back(arc);
	this->e++;
}

/* deletes arcs from heap memory*/
void MeanPayoffGame::delete_arcs(){
	for(unsigned int i=0; i < this->n_0 + this->n_1; i++){
		list<t_w_arc*>* arc_list = get_arcs(i);
		while(!arc_list->empty()){
			t_w_arc* arc = arc_list->front();
			arc_list->pop_front();
			delete arc;
		}
	}
	delete [] this->arcs;
}

/* checks whether @this is not empty and 
every vertex has at least one outgoing neighbour */
bool MeanPayoffGame::is_well_defined(){
	if(this->n_0 + this->n_1 == 0) return false; // do not consider empty games
	for(unsigned int i=0; i < this->n_0 + this->n_1; i++)
		if(get_arcs(i)->empty()) return false;
	return true; 	
}

/* decrements this->e */
void MeanPayoffGame::remove_arc(unsigned int u){
	get_arcs(u)->pop_front();
	this->e--;
}

/* prints to output */
void MeanPayoffGame::print(){
	o_stream << "## Mean Payoff Game definition: "<< endl;
	o_stream << this->n_0 << " " << this->n_1 << endl;
	o_stream << this->e << endl;
	for(unsigned int i=0; i < this->n_0 + this->n_1; i++){
		list<t_w_arc*>* arc_list = get_arcs(i);
		list<t_w_arc*>::iterator it = arc_list->begin();
		for (it; it != arc_list->end(); it++){
			t_w_arc* arc = *it;
			o_stream << i << " " << arc->head_idx << " " << arc->weight << endl;
		}
	}	
	o_stream << "## End of MPG definition." << endl;
}
