/*  
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "h/solver.h"

using namespace std;

struct t_graph{
	unsigned int n;
	std::list<t_w_arc*>* arcs;
};

/* internal solver function prototypes  */
bool is_vertex_winning(unsigned int vertex, int player, MeanPayoffGame* mpg);
bool is_vertex_optimal(unsigned int u, t_rational value, MeanPayoffGame* mpg );
bool is_all_winning(int player, MeanPayoffGame* mpg);
t_graph* compute_max_strategy_subgraph(MeanPayoffGame* mpg);
t_feasible_sched* compute_feasible_sched(t_graph* G, unsigned int n_0);
t_w_arc* binary_search_winning_arc(unsigned int u, int player, MeanPayoffGame* mpg);
t_w_arc* binary_search_optimal_arc(unsigned int u, t_rational value, MeanPayoffGame* mpg);
unsigned int add_fake_source(t_graph* G, unsigned int n_0);
bool arc_relax(unsigned int tail, bool* infinite_distance, t_w_arc* arc, t_feasible_sched* t);
void delete_graph(t_graph* G);
/**/

/* decides whether @htn is consistent or not */
bool is_consistent(HyperTemporalNetwork* htn){
	Reduction reduction;
	MeanPayoffGame *mpg = reduction.reduce2mpg(htn);
	bool consistent = is_all_winning(MAX, mpg);
	delete mpg;
	return consistent;
}

/* decides whether vertex u is 
   winning starting position for @player*/
bool is_vertex_winning(unsigned int u, int player, MeanPayoffGame* mpg){
	bool* decision = mpg->ZP_solve_decision();
	bool ret = true;
	if(player == MAX && decision[u]==false ||
	   player == MIN && decision[u]==true)
		ret=false;
	delete decision;
	return ret;
}

bool rational_equals(t_rational x, t_rational y){
	return x.num*y.denom == x.denom*y.num;
}

/* decides whether vertex @u has optimal value @value*/
bool is_vertex_optimal(unsigned int u, t_rational value, MeanPayoffGame* mpg){
	t_rational* values = mpg->ZP_solve_opt_values();
	t_rational opt_val = values[u];
	delete [] values;
	return rational_equals(opt_val, value);
}

/* decides whether all vertices are 
   winning starting position for @player*/
bool is_all_winning(int player, MeanPayoffGame* mpg){
	bool* decision = mpg->ZP_solve_decision();
	bool ret = true;
	unsigned int size = mpg->get_n_0() + mpg->get_n_1();
	for(unsigned int u=0; u < size; u++){
		if(player == MAX && decision[u]==false ||
		   player == MIN && decision[u]==true){
			ret = false; break;
		}
	}
	delete decision;
	return ret;
}

/* computes and returns a feasible scheduling for @htn */
/* pre-cond: assume @htn is consistent */
t_feasible_sched* compute_feasible_sched(HyperTemporalNetwork* htn){
	Reduction reduction;
	MeanPayoffGame *mpg = reduction.reduce2mpg(htn);
	t_graph* G = compute_max_strategy_subgraph(mpg);
	unsigned int s_idx = add_fake_source(G, mpg->get_n_0());
	t_feasible_sched* htn_sched = compute_feasible_sched(G, mpg->get_n_0());	
	delete_graph(G);
	delete mpg;
	return htn_sched;
}

/* computes and returns a strategy-induced subgraph G_{\pi} */
/* pre-cond: assume every vertex of mpg is winning for Max */
t_graph* compute_max_strategy_subgraph(MeanPayoffGame* mpg){
	// init graph
	t_graph* G = new t_graph;
	unsigned int size = mpg->get_n_0() + mpg->get_n_1();
	G->n = size;
	G->arcs = new list<t_w_arc*>[size];
	// every arc out of Min vertex
	for(unsigned int u=0; u < mpg->get_n_0(); u++){
		list<t_w_arc*>* arcs = mpg->get_arcs(u);
		G->arcs[u] = *arcs; // copy arcs
	}
	// !1 arc out of each Max vertex with ZP binary search	
	t_rational* values;
	if(CLASSICAL_ZP_MODE) values = mpg->ZP_solve_opt_values();
	for(unsigned int u = mpg->get_n_0(); u < size; u++){
		t_w_arc* winning_arc;
		if(CLASSICAL_ZP_MODE){
			t_rational value = values[u];
			winning_arc = binary_search_optimal_arc(u, value, mpg);
		}else
			winning_arc = binary_search_winning_arc(u, MAX, mpg);
		G->arcs[u].push_front(winning_arc);
	}
	delete [] values;
	return G;
}

t_w_arc* binary_search_optimal_arc(unsigned int u, t_rational value, MeanPayoffGame* mpg){
	list<t_w_arc*>* old_arc_list = mpg->get_arcs(u);
	if(old_arc_list->size() == 1) return old_arc_list->front();
	list<t_w_arc*> first_half;
	list<t_w_arc*> second_half;
	list<t_w_arc*>::iterator it = old_arc_list->begin();
	unsigned int i=0; // fill first and second half
	for(it; it != old_arc_list->end(); it++, i++){
		t_w_arc* arc = *it;
		if(i < old_arc_list->size()/2)
			first_half.push_back(arc);
		else
			second_half.push_back(arc);
	}
	t_w_arc* arc;
	mpg->set_arcs(u, &first_half);
	if(is_vertex_optimal(u, value, mpg)) // go recursively of first half
		arc = binary_search_optimal_arc(u, value, mpg);
	else{ // otherwise, go recursively on second half
		mpg->set_arcs(u, &second_half);
		arc = binary_search_optimal_arc(u, value, mpg);
	}
	mpg->set_arcs(u, old_arc_list); // restore old arcs 
	return arc;
}

/* returns a winning arc out of @u for @player */
/* pre-cond: assume u is winning for player "int player" */
t_w_arc* binary_search_winning_arc(unsigned int u, int player, MeanPayoffGame* mpg){
	list<t_w_arc*>* old_arc_list = mpg->get_arcs(u);
	if(old_arc_list->size() == 1) return old_arc_list->front();
	list<t_w_arc*> first_half;
	list<t_w_arc*> second_half;
	list<t_w_arc*>::iterator it = old_arc_list->begin();
	unsigned int i=0; // fill first and second half
	for(it; it != old_arc_list->end(); it++, i++){
		t_w_arc* arc = *it;
		if(i < old_arc_list->size()/2)
			first_half.push_back(arc);
		else
			second_half.push_back(arc);
	}
	t_w_arc* arc;
	mpg->set_arcs(u, &first_half);
	if(is_vertex_winning(u, player, mpg)) // go recursively of first half
		arc = binary_search_winning_arc(u, player, mpg);
	else{ // otherwise, go recursively on second half
		mpg->set_arcs(u, &second_half);
		arc = binary_search_winning_arc(u, player, mpg);
	}
	mpg->set_arcs(u, old_arc_list); // restore old arcs 
	return arc;
}

/* adds a fake source vertex to V_G, connecting it to black vertices 
with new zero-weighted arcs */
unsigned int add_fake_source(t_graph* G, unsigned int n_0){
	unsigned int src_idx = G->n;
	G->n++;
	list<t_w_arc*>* new_arc_array = new list<t_w_arc*>[G->n];
	for(int u=0; u < G->n-1; u++){
		new_arc_array[u] = G->arcs[u]; // copy arc list
		if(u < n_0){ // add arc from fake source to u
			t_w_arc* new_arc = new t_w_arc;
			new_arc->head_idx = u;
			new_arc->weight = 0;
			new_arc_array[src_idx].push_back(new_arc);
		}
	}
	delete [] G->arcs;
	G->arcs = new_arc_array;		
}

/* computes and returns a feasible scheduling for @G */
/* pre-cond: assume G is conservative with source idx G->n-1 */
/* this is a Bellman-Ford based procedure*/
t_feasible_sched* compute_feasible_sched(t_graph* G, unsigned int n_0){
	t_feasible_sched t;
	t.size = G->n;
	t.p = new long[t.size];
	bool infinite_distance[t.size];
	t.p[t.size-1] = 0;
	infinite_distance[t.size-1] = false;
	for(unsigned int u; u < t.size-1; u++)
		infinite_distance[u] = true;
	for(unsigned int phase=0; phase < G->n-1; phase++){
		for(unsigned int u=0; u < G->n; u++){
			list<t_w_arc*>* arc_list = &G->arcs[u];
			list<t_w_arc*>::iterator it = arc_list->begin();
			for(it; it != arc_list->end(); it++){
				t_w_arc* arc = *it;
				if(arc_relax(u, infinite_distance, arc, &t)){
					t.p[arc->head_idx] = t.p[u] + arc->weight;
					infinite_distance[arc->head_idx] = false;
				}
			}
		}
	}
	t_feasible_sched* htn_sched = new t_feasible_sched;
	htn_sched->size = n_0;
	htn_sched->p = new long[n_0];
	for(unsigned int u=0; u < n_0; u++)
		htn_sched->p[u] = - t.p[u]; // revert sign
	delete [] t.p;
	return htn_sched;
}

/* computes a negative cycle of @htn */
/* pre-cond: assume htn is not consistent */
t_neg_cycle* compute_neg_cycle(HyperTemporalNetwork* htn){
	Reduction reduction;
	MeanPayoffGame *mpg = reduction.reduce2mpg(htn);
	mpg->print();
	bool* decision = mpg->ZP_solve_decision();
	t_neg_cycle* neg_cycle = new t_neg_cycle;
	for(unsigned int u=0; u < mpg->get_n_0(); u++)
		if(decision[u]==false){ // u is winning for MIN
			t_w_arc* arc = binary_search_winning_arc(u, MIN, mpg);
			unsigned int hyperarc_idx = arc->head_idx - mpg->get_n_0();
			list<t_hyperarc*>* hyper_arcs = htn->get_hyperarcs(u);
			list<t_hyperarc*>::iterator it = hyper_arcs->begin();
			for(it; it != hyper_arcs->end(); it++){
				t_hyperarc* hyperarc = *it;
				if(hyperarc->idx == hyperarc_idx){
					pair<unsigned int, t_hyperarc*> entry(u, hyperarc);
					neg_cycle->SC.insert(entry); // insert hyperarc 
					break;
				}
			}
		}
	delete [] decision;
	delete mpg;
	return neg_cycle;
}

/* relaxes @arc with respect to its @tail and distance function @t*/
bool arc_relax(unsigned int tail, bool* infinite_distance, t_w_arc* arc, t_feasible_sched* t){
	if(infinite_distance[tail]) return false;
	long candidate_distance = t->p[tail] + arc->weight;
	return (infinite_distance[arc->head_idx] || 
		candidate_distance < t->p[arc->head_idx]) ? true : false;
}

/* deletes graph G from heap memory */
void delete_graph(t_graph* G){
	delete [] G->arcs;
	delete G;
}
