/*  
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/******************************************************
** Source Code for TIME2014 - Consistency-HTN Tool   **
** Author: Carlo Comin, 03/2014		 	     **
*******************************************************/

#include <iostream>
#include "time.h"
#include "h/conf.h"
#include "h/htn.h"
#include "h/mpg.h"
#include "h/reduction.h"
#include "h/solver.h"

using namespace std;

const char* CONSISTENT = "consistent.";
const char* INCONSISTENT = "inconsistent.";
const char* INPUT_FILE_MPG="data/mpg.dat";
const char* INPUT_FILE_HTN_CONSISTENT="data/htn_consistent.dat";
const char* INPUT_FILE_HTN_INCONSISTENT="data/htn_inconsistent.dat";
const char* OUTPUT_FILE="data/output.dat";
const bool VERBOSE_MODE=false;
bool CLASSICAL_ZP_MODE = false;
bool CHECK_CONSISTENCY = false;
bool COMPUTE_CERTIFICATES = false;

ofstream o_stream;
/*********************************************
* Prototypes
*********************************************/
void test1();
void test2();
void test3();
void test4();
void test5();
void test6();
void pre_args(int argc, char** argv); 
void print_feasible_sched(t_feasible_sched* t);
void print_neg_cycle(t_neg_cycle* c);
clock_t pre_test(int n);
void post_test(int n, clock_t t_start);

void welcome_msg();
/********************************************
** TIME2014 - Consistency-HTN Tool main()
*********************************************/
int main(int argc, char** argv){
	o_stream.open(OUTPUT_FILE);
	pre_args(argc, argv);
	welcome_msg();
	/*test1();
	test2();
	test3();
	test4();
	test5();*/
	test6();
	o_stream.close();
	return 0;
} 

/*******************
** default test
*******************/
void default_htn_test(HyperTemporalNetwork* htn){
	htn->print();
	cout << "\tDeciding HTN: ... " << endl;
	bool consistent = CHECK_CONSISTENCY ? is_consistent(htn) : true;
	string str_decision = consistent ? CONSISTENT : INCONSISTENT; 
	cout << "\t\tHTN is " << str_decision << endl;
	o_stream << "\t\tHTN is " << str_decision << endl; 
	if(COMPUTE_CERTIFICATES)
		if(consistent){
			cout << "\tComputing a feasible scheduling... "<<endl;
			t_feasible_sched* htn_sched = compute_feasible_sched(htn);
			//htn->assert_feasible_sched(htn_sched); // check htn_sched 
			print_feasible_sched(htn_sched);	
			delete [] htn_sched->p;
			delete htn_sched;
			cout << "\tFeasible scheduling computed!"<<endl;
		} else{
			cout << "\tComputing a negative cycle... "<<endl;
			t_neg_cycle* htn_neg_cyc = compute_neg_cycle(htn);
			//htn->assert_neg_cycle(htn_neg_cyc); // check htn_neg_cyc
			print_neg_cycle(htn_neg_cyc);	
			delete htn_neg_cyc;
			cout << "\tNegative cycle computed!"<<endl;
		}
	delete htn;
}



/********************
** test 1: htn2mpg
********************/
void test1(){
	clock_t t_start = pre_test(1);
	HyperTemporalNetwork* htn = new HyperTemporalNetwork(INPUT_FILE_HTN_CONSISTENT);
	default_htn_test(htn);
	post_test(1, t_start);
}

/*******************
** test 2: htn2mpg
********************/
void test2(){
	clock_t t_start = pre_test(2);
	HyperTemporalNetwork* htn = new HyperTemporalNetwork(INPUT_FILE_HTN_INCONSISTENT);
	default_htn_test(htn);
	post_test(2, t_start);
}

/********************
** test 3 mpg2htn
********************/
void test3(){
	clock_t t_start = pre_test(3);
	MeanPayoffGame* mpg = new MeanPayoffGame(INPUT_FILE_MPG);
	Reduction reduction;
	HyperTemporalNetwork* htn = reduction.reduce2htn(mpg);
	default_htn_test(htn);
	delete mpg;
	post_test(3, t_start);
}

/***********************
** test 4 mpg2htn 
***********************/
void test4(){
	clock_t t_start = pre_test(4); 
	MeanPayoffGame* mpg = new MeanPayoffGame("data/my_mpg1.dat");
	Reduction reduction;
	HyperTemporalNetwork* htn = reduction.reduce2htn(mpg);
	default_htn_test(htn);
	delete mpg;
	post_test(4, t_start);
}

/***********************
** test 5 mpg2htn
***********************/
void test5(){
	clock_t t_start = pre_test(5);
	MeanPayoffGame* mpg = new MeanPayoffGame("data/my_mpg2.dat");
	Reduction reduction;
	HyperTemporalNetwork* htn = reduction.reduce2htn(mpg);
	default_htn_test(htn);
	delete mpg;
	post_test(5, t_start);
}

/***********************
** test 6 mpg2htn
***********************/
void test6(){
	clock_t t_start = pre_test(6);
	MeanPayoffGame* mpg = new MeanPayoffGame("data/pg_mpg.dat");
	Reduction reduction;
	HyperTemporalNetwork* htn = reduction.reduce2htn(mpg);
	default_htn_test(htn);
	delete mpg;
	post_test(6, t_start);
}

clock_t pre_test(int n){
       	clock_t start = clock();
	cout << " Executing test" << n << "...\n\tWriting output data to " 
		<< OUTPUT_FILE << "..." << endl;
	o_stream << " Executing test" << n << "...\n\tWriting output data to " 
		<< OUTPUT_FILE << "..." << endl; 
	return start;
}

void post_test(int n, clock_t t_start){
	cout << " Test" << n << "completed!" << endl;
	o_stream << " Test" << n << "completed!" << endl;
	clock_t t_end = clock();
	double sec_taken = (double)(t_end - t_start)/CLOCKS_PER_SEC;
	cout << " Execution time: " << sec_taken << " sec" << endl << endl;
	o_stream << " Execution time: " << sec_taken << " sec" << endl << endl;
}

/* prints to output streams the scheduling @htn_sched  */
void print_feasible_sched(t_feasible_sched* htn_sched){
	cout << "\tFeasible scheduling:" << endl;
	o_stream << "\tFeasible scheduling:" << endl; 
	for(unsigned int u=0; u < htn_sched->size; u++){
		cout << "\t\tHTN's Node #" << u 
			<< " scheduled at timepoint: " << htn_sched->p[u] << endl; 
		o_stream << "\t\tNode #" << u 
			<< " scheduled at timepoint: " << htn_sched->p[u] << endl;  
	}
}

/* prints to output streams the negative cycle @neg_cyc */
void print_neg_cycle(t_neg_cycle* neg_cyc){
	cout << "\tNegative cycle:" << endl;
	o_stream << "\tNegative cycle:" << endl;
	map<unsigned int, t_hyperarc*>::iterator it = neg_cyc->SC.begin();
	for(it; it != neg_cyc->SC.end(); it++){
		t_hyperarc* hyperarc = it->second;
		map<unsigned int, long>::iterator it2 = hyperarc->weighted_tail.begin();
		cout << "\t\tHTN's Hyperarc #" 
			<< hyperarc->idx << ": head = " 
			<< hyperarc->head << " tail = ";
		o_stream << "\t\tHTN's Hyperarc #" 
			<< hyperarc->idx << ": head = " 
			<< hyperarc->head << " tail = ";
		for(it2; it2 != hyperarc->weighted_tail.end(); it2++){
			cout << " " << it2->first;	
			o_stream << " " << it2->first;
		}
		cout << endl; o_stream << endl;
	}
}

void pre_args(int argc, char** argv){
	if(argc > 4) throw "arguments exception";
	for(int i=1; i < argc; i++){
		string opt = argv[i];
		if(opt=="--classical_mode")
			CLASSICAL_ZP_MODE = true;
		if(opt=="--check_consistency")
			CHECK_CONSISTENCY = true;
		if(opt=="--compute_certificates")
			COMPUTE_CERTIFICATES = true;
	}
}

/* prints to output a welcome message */
void welcome_msg(){
	cout << endl;
	cout << "*                                                     *"<<endl;
	cout << "*******************************************************"<<endl;
	cout << "  **          ---- TIME2014's Code ----            **"<<endl;
	cout << "  **   Hyper Temporal Network's Consistency Tool   **"<<endl;
	cout << "*******************************************************"<<endl;
	cout << "*                                                     *"<<endl;
	cout << "*                         _____                       *"<<endl;
	cout << "*                      _.'_____`._                    *"<<endl;
	cout << "*                    . .-  12  -. .                   *"<<endl;
	cout << "*                   /,' 11      1 `.\'                 *"<<endl;                 
	cout << "*                  // 10      /   2  \\                *"<<endl;
	cout << "*                 ;;         /       ::               *"<<endl;
	cout << "*                 || 9  ----O      3 ||               *"<<endl;
	cout << "*                 ::                 ;;               *"<<endl;
	cout << "*                  \\ 8            4 //                *"<<endl;
	cout << "*                   \\`. 7       5  ,/                 *"<<endl;
	cout << "*                    '.`-.__6__.-'.'                  *"<<endl;
	cout << "*                     ((-._____.-))                   *"<<endl;
	cout << "**                    _))       ((_                  **"<<endl;
	cout << "* **                 '--'       '--'               * **"<<endl;
	cout << endl;
}
